// C/C++ headers

#include <stdio.h>

// QALE headers

#include "AttributeVisitor.h"
#include "AttributeNode.h"
#include "Widget.h"
#include "MessageReceiver.h"
#include "Component.h"

/*****************************************************************************
 * Definition of class: AttributeVisitor
 *****************************************************************************/

AttributeVisitor::AttributeVisitor(MessageReceiver &receiver)
   : mReceiver(receiver)
{
   // Nothing to do
}

AttributeVisitor::~AttributeVisitor()
{
   // Nothing to do
}

void AttributeVisitor::visited(AttributeNode *attributeNode)
{
   string errMsg                      = "";
   char *qualifiedName                = attributeNode->getQualifiedName();
   AttributeNode::Attribute attribute = attributeNode->get();
   Component *theComponent            = NULL;

   if(*qualifiedName)
   {
      Widget *theWidget = findWidget(qualifiedName);

      if(theWidget != NULL)
         theComponent = theWidget->getComponent();
   }
   else
   {
      Widget *theWidget = getCurrentWidget();

      // Without qualifications the attributes "dw" and "dh" refer
      // to the current widget's default width and default height.
      // Without qualifications the attributes: l, r, t, b, x, y, w, h
      // refer to our parent's attributes.

      if(attribute != AttributeNode::_DEFAULT_WIDTH &&
         attribute != AttributeNode::_DEFAULT_HEIGHT)
      {
         theWidget = theWidget->getParent();
      }

      if(theWidget != NULL)
         theComponent = theWidget->getComponent();
   }

   if(theComponent == NULL)
   {
      // Couldn't find the appropriate widget

      errMsg = getErrorMessage("undefined widget attribute", attributeNode);
      mReceiver.accept(errMsg.c_str());
   }
   else
   {
      // Lastly we have to enforce some ordering rules.  Since our widgets
      // are sized/located in a particular order it is a "bad thing" (tm)
      // for a widget to rely on an attribute of another widget if that
      // attribute won't be set until after the first widget has been
      // processed.  Our rules are:
      //    (a) dw and dh are order independent
      //    (b) all sizes are fully calculated before any locations are
      //        calculated so any sizing related attribute can only
      //        refer to sizing related attributes of widgets whose
      //        index # is lower
      //    (c) a location attribute can refer to any other widget's
      //        sizing information but can only refer to the location
      //        attributes of a widget whose index # is lower

      int error = 0;

      if(attribute != AttributeNode::_DEFAULT_WIDTH &&
         attribute != AttributeNode::_DEFAULT_HEIGHT)
      {
         if(getField() == SIZE_WIDTH || getField() == SIZE_HEIGHT)
         {
            switch(attribute)
            {
               case AttributeNode::_LEFT   :
               case AttributeNode::_RIGHT  :
               case AttributeNode::_TOP    :
               case AttributeNode::_BOTTOM :
               case AttributeNode::_XCENTER:
               case AttributeNode::_YCENTER:
               {
                  // A widget's height and/or width shouldn't depend
                  // upon another widget's positioning

                  error = 1;
               }
               break;

               default:
               {
                  // We're depending upon another widget's height or width
                  // but that widget won't be processed until after we
                  // are

                  if(theComponent->getIndex() >=
                     getCurrentWidget()->getComponent()->getIndex())
                        error = 1;
               }
               break;
            }
         }
         else
         {
            switch(attribute)
            {
               case AttributeNode::_LEFT   :
               case AttributeNode::_RIGHT  :
               case AttributeNode::_TOP    :
               case AttributeNode::_BOTTOM :
               case AttributeNode::_XCENTER:
               case AttributeNode::_YCENTER:
               {
                  // We're depending upon the location of another widget
                  // but that widget won't be positioned until after we
                  // are

                  if(theComponent->getIndex() >=
                     getCurrentWidget()->getComponent()->getIndex())
                        error = 1;
               }
               break;

               default:
               {
                  // No problem for _WIDTH and _HEIGHT
               }
               break;
            }
         }
      }

      if(error)
      {
         errMsg = getErrorMessage(
            "forward reference to widget attribute", attributeNode
         );

         mReceiver.accept(errMsg.c_str());
      }
      else
      {
         attributeNode->setComponent(theComponent);
      }
   }
}

void AttributeVisitor::visited(BinaryNode *)   {}
void AttributeVisitor::visited(ConstantNode *) {}
void AttributeVisitor::visited(DefineNode *)   {}
void AttributeVisitor::visited(PropertyNode *) {}
void AttributeVisitor::visited(MinMaxNode *)   {}
void AttributeVisitor::visited(UnaryNode *)    {}
