%{

#include <string.h>

#include "AttributeNode.h"
#include "Alignment.h"
#include "PointerSet.h"
#include "qale.yacc.h"
#include "QALELexer.h"

/*****************************************************************************
 * Definitions for class "QALELexer"
 *****************************************************************************/

QALELexer::QALELexer()
   :
   mFileOffset(0),
   mLineNumber(1),
   mLineOffset(0),
   mText(0),
   mTextIndex(0),
   mTextLength(0)
{}

QALELexer *QALELexer::getInstance()
{
   static QALELexer theEnvironment;

   return(&theEnvironment);
}

void QALELexer::begin(const char *text)
{
   YY_NEW_FILE;

   int length = strlen(text);

   mFileOffset = 0;
   mLineNumber = 1;
   mLineOffset = 0;
   mText       = new char[length];
   mTextIndex  = 0;
   mTextLength = length;

   memcpy(mText, text, mTextLength);
}

void QALELexer::end()
{
   delete [] mText;
}

int QALELexer::getLineNumber()
{
   return(mLineNumber);
}

void QALELexer::nextLine()
{
   ++mLineNumber;
   mLineOffset = 0;
}

int QALELexer::getText(char *dest, int maxSize)
{
   int bytesRemaining = mTextLength - mTextIndex;
   int count          = bytesRemaining > maxSize ? maxSize : bytesRemaining;

   memcpy(dest, mText, count);

   mTextIndex += count;

   return(count);
}

int QALELexer::getFileOffset()
{
   return(mFileOffset);
}

int QALELexer::getLineOffset()
{
   return(mLineOffset);
}

void QALELexer::getCurrentLine(char *line)
{
   // Grab the current line until we reach a \n or MAX_LINE_LENGTH
   // characters, whichever is smaller

   int j = getFileOffset() - getLineOffset();
   int i = 0;

   for(i = 0; j < mTextLength && i < MAX_LINE_LENGTH; i ++, j ++)
   {
      if(mText[j] == '\n')
         break;

      line[i] = mText[j];
   }

   line[i] = '\0';
}

void QALELexer::advance(char *theText)
{
   int nChars = strlen(theText);

   mFileOffset += nChars;
   mLineOffset += nChars;
}

/*****************************************************************************
 * Local function prototypes
 *****************************************************************************/

static int my_yyinput(char *buffer, int maxsize);

extern "C"
{
   int yylex(YYSTYPE *yylval, PointerSet *p);
}

#undef YY_INPUT

#define YY_INPUT(buffer, count, maxsize) (count = my_yyinput(buffer, maxsize))

#define YY_DECL int yylex(YYSTYPE *yylval, PointerSet *p)

// Define this symbol so that the generated
// source does not contain a reference to "isatty" (a unix-specific function)

#define YY_ALWAYS_INTERACTIVE (1)

%}

real                 ([0-9]+|[0-9]*\.[0-9]+|[0-9]+\.[0-9]*)([eE]([+-]?)[0-9]+)?
identifier           [a-zA-Z_][a-zA-Z0-9_]*
parent               \[[pP][aA][rR][eE][nN][tT]\]
this                 \[[tT][hH][iI][sS]\]
extended_identifier  ({parent}|{this}|{identifier})
qualified_identifier {extended_identifier}(\.{extended_identifier})*

   /* -- Use flex's "exclusive start states" feature to avoid having to
         prefix <INITIAL> to all tokens that should be recognized
         while in the default state, note "%x" rather than "%s" -- */

%x STATE_MULTI_LINE_COMMENT
%x STATE_SINGLE_LINE_COMMENT

%%

   /* -- Token handling for the default state -- */

{real} {
   yylval->dvalue = atof(yytext);
   QALELexer::getInstance()->advance(yytext);
   return(REAL);
}

   /* -- Tokens related to attributes -- */

[lL] {
      yylval->theAttribute = AttributeNode::_LEFT;
      QALELexer::getInstance()->advance(yytext);
      return(LEFT);
}

[rR] {
      yylval->theAttribute = AttributeNode::_RIGHT;
      QALELexer::getInstance()->advance(yytext);
      return(RIGHT);
}

[tT] {
      yylval->theAttribute = AttributeNode::_TOP;
      QALELexer::getInstance()->advance(yytext);
      return(TOP);
}

[bB] {
      yylval->theAttribute = AttributeNode::_BOTTOM;
      QALELexer::getInstance()->advance(yytext);
      return(BOTTOM);
}

[xX] {
      yylval->theAttribute = AttributeNode::_XCENTER;
      QALELexer::getInstance()->advance(yytext);
      return(XCENTER);
}

[yY] {
      yylval->theAttribute = AttributeNode::_YCENTER;
      QALELexer::getInstance()->advance(yytext);
      return(YCENTER);
}

[wW] {
      yylval->theAttribute = AttributeNode::_WIDTH;
      QALELexer::getInstance()->advance(yytext);
      return(WIDTH);
}

[hH] {
      yylval->theAttribute = AttributeNode::_HEIGHT;
      QALELexer::getInstance()->advance(yytext);
      return(HEIGHT);
}

[dD][wW] {
      yylval->theAttribute = AttributeNode::_DEFAULT_WIDTH;
      QALELexer::getInstance()->advance(yytext);
      return(DEFAULT_WIDTH);
}

[dD][hH] {
      yylval->theAttribute = AttributeNode::_DEFAULT_HEIGHT;
      QALELexer::getInstance()->advance(yytext);
      return(DEFAULT_HEIGHT);
}

   /* -- Tokens related to alignments -- */

[lL][tT] {
      yylval->theType = Alignment::_LT;
      QALELexer::getInstance()->advance(yytext);
      return(LT);
}

[lL][cC] {
      yylval->theType = Alignment::_LC;
      QALELexer::getInstance()->advance(yytext);
      return(LC);
}

[lL][bB] {
      yylval->theType = Alignment::_LB;
      QALELexer::getInstance()->advance(yytext);
      return(LB);
}

[cC][tT] {
      yylval->theType = Alignment::_CT;
      QALELexer::getInstance()->advance(yytext);
      return(LT);
}

[cC][cC] {
      yylval->theType = Alignment::_CC;
      QALELexer::getInstance()->advance(yytext);
      return(LC);
}

[cC][bB] {
      yylval->theType = Alignment::_CB;
      QALELexer::getInstance()->advance(yytext);
      return(LB);
}

[rR][tT] {
      yylval->theType = Alignment::_RT;
      QALELexer::getInstance()->advance(yytext);
      return(LT);
}

[rR][cC] {
      yylval->theType = Alignment::_RC;
      QALELexer::getInstance()->advance(yytext);
      return(LC);
}

[rR][bB] {
      yylval->theType = Alignment::_RB;
      QALELexer::getInstance()->advance(yytext);
      return(LB);
}

   /* -- Additional specific tokens -- */

[dD][eE][fF][iI][nN][eE] {
      QALELexer::getInstance()->advance(yytext);
      return(DEFINE);
}

[mM][iI][nN] {
   QALELexer::getInstance()->advance(yytext);
   return(MIN);
}

[mM][aA][xX] {
   QALELexer::getInstance()->advance(yytext);
   return(MAX);
}

   /**************************************************************************
    * For both "identifier" and "qualified_identifier" there is a
    * possibility of a memory leak if we're not careful.  We have
    * to allocate space for "yytext" each time but if we encounter
    * a syntax error the corresponding "delete []" code in qale.yacc will
    * never be executed.
    *
    * The way we handle this is by recording the address of each thing we
    * allocate inside "p" (a PointerSet object) if we find out way to
    * qale.yacc (no syntax error) we'll delete the "identifier" or
    * "qualified identifier" we allocated and remove its pointer from
    * our PointerSet.  If we don't make it back to qale.yacc PointerSet's
    * destructor will delete the pointer for us.
    **************************************************************************/

{identifier} {

   yylval->identifier = strcpy(
      new char[strlen(yytext) + 1],
      yytext
   );

   p->add(yylval->identifier);

   QALELexer::getInstance()->advance(yytext);
   return(IDENTIFIER);
}

{qualified_identifier} {

   yylval->qualified_identifier = strcpy(
      new char[strlen(yytext) + 1],
      yytext
   );

   p->add(yylval->qualified_identifier);

   QALELexer::getInstance()->advance(yytext);
   return(QUALIFIED_IDENTIFIER);
}

[ \t\r] {
   /* -- Ignore whitespace -- */
   QALELexer::getInstance()->advance(yytext);
}

\n {
   QALELexer::getInstance()->advance(yytext);
   QALELexer::getInstance()->nextLine();
}
      
"/*" {
   QALELexer::getInstance()->advance(yytext);
   BEGIN STATE_MULTI_LINE_COMMENT;
}

"//" {
   QALELexer::getInstance()->advance(yytext);
   BEGIN STATE_SINGLE_LINE_COMMENT;
}

"#" {
   QALELexer::getInstance()->advance(yytext);
   BEGIN STATE_SINGLE_LINE_COMMENT;
}

. {
   QALELexer::getInstance()->advance(yytext);
   return(*yytext);
}

   /* -- Token handling for STATE_MULTI_LINE_COMMENT  -- */

<STATE_MULTI_LINE_COMMENT>\n {
   QALELexer::getInstance()->advance(yytext);
   QALELexer::getInstance()->nextLine();
}

<STATE_MULTI_LINE_COMMENT>. {
   /* -- Ignore -- */
   QALELexer::getInstance()->advance(yytext);
}

<STATE_MULTI_LINE_COMMENT>"*/" {
   QALELexer::getInstance()->advance(yytext);
   BEGIN INITIAL;
}

   /* -- Token handling for STATE_SINGLE_LINE_COMMENT  -- */

<STATE_SINGLE_LINE_COMMENT>\n {
   QALELexer::getInstance()->advance(yytext);
   QALELexer::getInstance()->nextLine();
   BEGIN INITIAL;
}

<STATE_SINGLE_LINE_COMMENT>. {
   /* -- Ignore -- */
   QALELexer::getInstance()->advance(yytext);
}

%%

int yywrap() { return(1); }

static int my_yyinput(char *buffer, int maxsize)
{
   return(
      QALELexer::getInstance()->getText(buffer, maxsize)
   );
}
